<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2021 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @author     Malin Lorch <mlorch@liquidweb.com>
 * @date       Sep 23, 2021
 */

/**
 * @package InterWorx
 * @subpackage CLI
 */
class AcronisIntegrationTests extends CLI
{
    const INPUT_PASSWORD = 'password';
    const INPUT_USERNAME = 'username';

    const FUNCTION_TEST_ENABLE = 'testEnablePlugin';
    const FUNCTION_TEST = 'test';
    const FUNCTION_IS_READY = 'isReady';

    const SUCCESS_MESSAGE = 'Tests are done! Checkout log for results: ~iworx/plugins/acronis/log/phpunit.xml';
    const FAIL_TRY_LATER = 'System does not seem to be ready for testing yet. Try again later.';
    /**
     * Register functions.
     */
    public function registerOptArgs()
    {
        $this->_func(self::FUNCTION_TEST_ENABLE, 'run tests against system without plugin enabled. Credentials required. OBS - will take up to 2 hours if backup agent not installed');
        $this->_func(self::FUNCTION_TEST, 'run tests against system with plugin enabled. OBS - will take up to 2 hours if backup agent not installed');
        $this->_func(self::FUNCTION_IS_READY, 'Check if system is ready for integration tests.');

        $this->_param(self::INPUT_PASSWORD . '=s', 'Acronis Password');
        $this->_param(self::INPUT_USERNAME . '=s', 'Acronis Username');
    }

    /**
     * Runs test suites against enabled plugin.
     *
     * OBS - Can take up to 2 hours to execute on fresh installs without backup agent installed
     */
    protected function _test()
    {
        $this->_installBackupAgent();
        $this->_syncAgentScripts();

        if (!$this->_readyForTests()) {
            $this->fail(self::FAIL_TRY_LATER);
        }

        $this->_runTests();
        $this->success(self::SUCCESS_MESSAGE);
    }

    /**
     * Enables plugin & runs test suite.
     *
     * OBS - Can take up to 2 hours to execute on fresh installs without backup agent installed
     */
    protected function _testEnablePlugin()
    {
        $this->_requires(self::FUNCTION_TEST_ENABLE, self::INPUT_PASSWORD, self::INPUT_USERNAME);

        if ($this->_pluginEnabled()) {
            $this->fail('Plugin already enabled');
        }

        $this->_enablePlugin();
        $this->_installBackupAgent();
        $this->_syncAgentScripts();
        $this->_activatePlan();

        if (!$this->_readyForTests()) {
            $this->fail(self::FAIL_TRY_LATER);
        }

        $this->_runTests();
        $this->success(self::SUCCESS_MESSAGE);
    }

    /**
     * Install backup agent, activate plan, and wait for one hour.
     * Wait is needed for acronis api to perform less error prone.
     */
    private function _installBackupAgent()
    {
        if (!$this->_backupAgentInstalled()) {
            $this->_sendAcronisCtrlReq('installBackupAgent');
        }
    }

    /**
     * Enable plugin.
     */
    private function _enablePlugin()
    {
        require_once(__DIR__ . '/../Plugin/Acronis.php');

        $username = $this->_getOption(self::INPUT_USERNAME);
        $password = $this->_getOption(self::INPUT_PASSWORD);
        $url = 'https://us5-cloud.acronis.com';

        $args = array(
            'plugin_name' => 'acronis',
            'status' => 1,
            Plugin_Acronis::INPUT_PASSWORD => $password,
            Plugin_Acronis::INPUT_USERNAME => $username,
            Plugin_Acronis::INPUT_ACRONIS_URL => $url
        );

        $Reply = IW::FC()->routeFromPHP(
            'Ctrl_Nodeworx_Plugins',
            'editCommit',
            $args,
            new NodeWorx(1)
        );

        if (!$Reply->wasSuccessful()) {
            $this->fail($Reply->getErrorsAsString());
        }

        // clear instance so that autoload
        // finds acronis files
        Plugin_Manager::clearInstance();
    }

    /**
     * @param string $action
     * @param array  $args
     */
    private function _sendAcronisCtrlReq($action, array $args = array())
    {
        $Reply = IW::FC()->routeFromPHP(
            'Ctrl_Nodeworx_Acronis',
            $action,
            $args,
            new NodeWorx(1)
        );

        if (!$Reply->wasSuccessful()) {
            $this->fail($Reply->getErrorsAsString());
        }

        return $Reply;
    }

    /**
     * Run tests unit/integration.
     */
    private function _runTests()
    {
        $Command = new Command('/home/interworx/bin/phpunit --configuration /home/interworx/plugins/acronis/TESTS/phpunit.xml');
        $Command->exec();

        if (!$Command->wasSuccessful()) {
            $this->fail($Command->getResultMessage());
        }
    }

    /**
     * @return boolean
     */
    private function _backupAgentInstalled()
    {
        require_once(__DIR__ . '/../Model/BackupAgent/BackupAgent.php');

        $BackupAgent = new Model_BackupAgent_BackupAgent();
        return $BackupAgent->isInstalled();
    }

    /**
     * Sync agent scripts.
     */
    private function _syncAgentScripts()
    {
        $cmd = '/home/interworx/plugins/acronis/scripts/dev/syncAgentScripts.sh';
        $Command = new Command($cmd);
        //$Command->runAsRoot();
        $Command->exec();

        if (!$Command->wasSuccessful()) {
            $this->fail('Command failed: ' . $cmd);
        }
    }

    /**
     * Check if system is ready for integration tests.
     *
     * Determines if ready by triggering backup & awaiting it.
     * If listBackups includes backup after 5 min, system should be ready.
     */
    protected function _isReady()
    {
        if ($this->_readyForTests()) {
            $this->success('READY!');
        } else {
            $this->fail('NOT READY :( try again later');
        }
    }

    /**
     * check if system is ready for tests. Checks if plugin enabled, backup agent installed,
     * and if runBackupCommitAction is working (sometimes not working, tests will fail then.)
     */
    private function _readyForTests()
    {
        if (!$this->_pluginEnabled()) {
            $this->fail('plugin not enabled.');
        }

        if (!$this->_backupAgentInstalled()) {
            $this->fail('Backup agent not installed.');
        }

        if (!$this->_backupsExists()) {
            $this->_activatePlan();
            $this->_runBackup();
            $this->_awaitBackup();
        }

        return $this->_backupsExists();
    }

    /**
     * Check if plugin is enabled.
     */
    private function _pluginEnabled()
    {
        $Plugin = IW::PluginManager()->getPlugin('acronis');
        return $Plugin->isEnabled();
    }

    /**
     * Activate plan. Needed to run backup.
     */
    private function _activatePlan()
    {
        if (!$this->_planActivated()) {
            $this->_sendAcronisCtrlReq('activatePlan');
        }
    }

    /**
     * Run backup.
     */
    private function _runBackup()
    {
        $Reply = $this->_sendAcronisCtrlReq('runBackup');

        if (!$Reply->wasSuccessful()) {
            $this->fail('unable to run backup: ' . $Reply->getErrorsAsString());
        }
    }

    /**
     * Run backup & await api.
     */
    private function _backupsExists()
    {
        $backups = $this->_getBackups();
        return !empty($backups);
    }

    /**
     * Get backups.
     */
    private function _getBackups()
    {
        $Reply = $this->_sendAcronisCtrlReq('listBackups');
        return $Reply->getPayloadData();
    }

    /**
     * Await acronis api to include the newly created backup.
     * Wait for max 5 minutes.
     */
    protected function _awaitBackup()
    {
        $maxAttempts = 5;
        $attempts = 0;

        while (!$this->_backupsExists() && $attempts < $maxAttempts) {
            sleep(60);
            $attempts++;
        }
    }

    /**
     * Check if backup plan is activated. Needed for running backups.
     */
    private function _planActivated()
    {
        require_once(__DIR__ . '/../Model/Server/Server.php');

        $Server = new Model_Server_Server();
        return !empty($Server->getAssignedPlans());
    }
}
