<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Malin Lorch <mlorch@liquidweb.com>
 * @date    Nov 26, 2019
 */

use function PHPSTORM_META\type;

/**
 * Nodeworx Controller for Acronis integration
 *
 * @package Interworx
 * @subpackage Plugin_Acronis
 */
class Ctrl_Nodeworx_Acronis extends Ctrl_Nodeworx_Plugin
{
    const ACTIVE_TAB = 'active_tab';
    const IS_AUTHORIZED = 'authorized';
    const INSTALL_BACKUP_AGENT = 'install_backup_agent';


    private $_Plugin;
    private $_ActivitiesLogger;

    /**
     * block actions if not master nodeworx user.
     *
     * @param string $action
     * @return boolean
     */
    public function isActionAllowed($action)
    {
        if (IW::NW()->isReseller() || !$this->_Plugin->isCurrentNwUserAllowed()) {
            $this->_blockAction('##LG_ACTION_NOT_PERMITTED##');
            return false;
        }

        if (Ini::get(Ini::IWORX, 'goiworx') == '0') {
            return true;
        }

        $parentCheck = parent::isActionAllowed($action);
        if ($parentCheck === true) {
            if (IW::NW()->isReseller() === true) {
                if ($this->_isAllowedAsReseller($action) === false) {
                    $this->_blockAction('##LG_ACTION_NOT_PERMITTED##');
                    return false;
                }
            }
        }
        return $parentCheck;
    }

    /**
     * Is allowed as reseller.
     *
     * @param string $action
     * @return boolean
     */
    protected function _isAllowedAsReseller($action)
    {
        return false;
    }

    /**
     * Init function. Final construct will run this method.
     */
    protected function _init()
    {
        $this->_Plugin = $this->_getPlugin();
    }


    /**
     * Index action.
     */
    public function indexAction()
    {
        // ONLY UNCOMMENT FOR DEBUG PURPOSES
        // $this->_Plugin->getCache()->deleteAll();

        $isAuthorized = $this->_Plugin->userIsAuthorized();

        $View = $this->getView();
        $View->assign(self::IS_AUTHORIZED, $isAuthorized);
        $View->assign('nw_user', true);

        if (!$isAuthorized) {
            $this->_displayFailure('_notAuthorized', AcronisReply::LANG_NOT_AUTHORIZED, AcronisReply::ERROR_API_INVALID_CREDENTIALS);
        }

        $BackupAgent = $this->_Plugin->getBackupAgent();

        $backupAgentNotInstalled = !$BackupAgent->isInstalled();
        $scriptsMissing = $BackupAgent->scriptsMissing();
        if ($isAuthorized && ($backupAgentNotInstalled || $scriptsMissing)) {
            $View->assign(self::INSTALL_BACKUP_AGENT, true);
        }
    }

    /**
     * Delete all cache. Note: use with caution, as it may cause errors.
     *
     * Useful if troubleshooting / testing
     */
    public function deleteCacheAction()
    {
        $this->_Plugin->getCache()->deleteAll();
        $this->_displaySuccess('index', 'Successfully deleted cache');
    }

    /**
     * Exists because index action can't refer to itself
     * when displayFailure is triggered.
     */
    private function _notAuthorizedAction()
    {
        $View = $this->getView();
        $View->assign(self::IS_AUTHORIZED, false);
    }

    /**
     * Install backup agent.
     *
     * @param Form_SubmitOnly $Form
     */
    public function installBackupAgentCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_installBackupAgent', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _installBackupAgent(Form_SubmitOnly $Form)
    {
        $BackupAgent = $this->_getAcronisPlugin()->getBackupAgent();

        if ($BackupAgent->isInstalled()) {
            $BackupAgent->addPrePostScriptsIfMissing();
        }

        if (!$BackupAgent->isInstalled()) {
            $BackupAgent->install();
        }

        $this->_displaySuccess('index', AcronisReply::LANG_INSTALL_SUCCESS);
    }

    /**
     * Uninstall backup agent.
     *
     * @param Form_SubmitOnly $Form
     */
    public function uninstallBackupAgentCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_uninstallBackupAgent', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _uninstallBackupAgent(Form_SubmitOnly $Form)
    {
        $BackupAgent = $this->_Plugin->getBackupAgent();
        $BackupAgent->uninstall();
        $this->_displaySuccess('index', AcronisReply::LANG_UNINTALL_SUCCESS);
    }

    /**
     * Get dashboard data.
     *
     * @param Form_SubmitOnly $Form
     */
    public function getDashboardDataCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_getDashboardData', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _getDashboardData(Form_SubmitOnly $Form)
    {

        $storageData = $this->_getStorageData();
        $planDetails =  $this->_getBackupPlanDetails();
        $backupStatus = $this->_getBackupStatusFromResource();

        $payloadData = $this->_mergeArrays(array($storageData, $planDetails, $backupStatus));
        $this->_displayPayloadAsJson($payloadData);
    }

    /**
     * merge multiple arrays into one array.
     *
     * @param array $arrayOfArrays
     * @return array
     */
    private function _mergeArrays(array $arrayOfArrays)
    {
        $result = array();

        foreach ($arrayOfArrays as $arr) {
            $result = array_merge($result, $arr);
        }

        return $result;
    }

    /**
     * get storage data.
     *
     * @return array
     */
    private function _getStorageData()
    {
        $Server = $this->_Plugin->getServer();
        return $Server->getStorageData();
    }

    /**
     * Get cloud storage usage data.
     *
     * @param Form_SubmitOnly $Form
     */
    public function getStorageSizeCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_displayStorageSize', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _displayStorageSize(Form_SubmitOnly $Form)
    {
        $storageData = $this->_getStorageData();
        $size = $storageData['size'];
        $Payload = new Payload($size);
        $this->_displayPayload($Payload);
    }

    /**
     * Get backup plan status (enabled/disabled) and backup plan schedule.
     *
     * @param Form_SubmitOnly $Form
     */
    public function getBackupPlanDetailsCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_displayBackupPlan', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    public function _displayBackupPlan(Form_SubmitOnly $Form)
    {
        $payloadData = $this->_getBackupPlanDetails();
        $this->_displayPayloadAsJson($payloadData);
    }

    /**
     * @return array
     */
    private function _getBackupPlanDetails()
    {
        $planDetails = array();

        if ($this->_isProtected()) {
            $BackupPlan = $this->_getCurrentBackupPlan();
            $planDetails = $BackupPlan->getPlanDetails();
        }

        return $planDetails;
    }

    /**
     * Get backup plan details.
     *
     * @return Model_Backups_BackupPlan
     */
    private function _getCurrentBackupPlan()
    {
        $planId = $this->_getCurrentPlanId();
        return new Model_Backups_BackupPlan($planId);
    }


    /**
     * Display last backup status, date, and time.
     *
     * @param Form_SubmitOnly $Form
     */
    public function getBackupActivitiesCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_displayBackupInfo', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _displayBackupInfo(Form_SubmitOnly $Form)
    {
        $backupInfo = $this->_getBackupStatusFromResource();
        $this->_displayPayloadAsJson($backupInfo);
    }


    /**
     * Get storage data.
     *
     * @return array
     */
    private function _getBackupStatusFromResource()
    {
        $Server = $this->_getServer();

        // only display last backup & status if next backup is scheduled.
        $result = $this->_isProtected() ? $Server->getBackupStatus() : array();
        return $result;
    }

    /**
     * @return Model_Server_Server
     */
    private function _getServer()
    {
        return new Model_Server_Server();
    }

    /**
     * Get redirect URL for one-time login. Link is valid for 30 seconds.
     *
     * @param Form_SubmitOnly $Form
     */
    public function redirectToBackupCloudCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_redirectToBackupCloud', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _redirectToBackupCloud(Form_SubmitOnly $Form)
    {
        $Auth = $this->_Plugin->getAuth();
        $loginLink = $Auth->getOneTimeLoginLink();
        $data = array('redirect' => $loginLink);
        $this->_displayPayloadAsJson($data);
    }

    /**
     * Revoke backup plan.
     *
     * @param Form_SubmitOnly $Form
     */
    public function revokePlanCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_revokePlan', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _revokePlan(Form_SubmitOnly $Form)
    {
        if (!$this->_isProtected()) {
            throw new Exception_Acronis_BackupPlan_NoPlanAssigned();
        }

        $this->_removeResourceFromPlan();

        $this->_awaitApiUpdateRevokePlan();
        $this->_displayPayload(new Payload('Backup plan was successfully revoked.'));
    }

    /**
     * Remove resource from backup plan.
     */
    private function _removeResourceFromPlan()
    {
        $currentPlanId = $this->_getCurrentPlanId();
        $BackupPlan = $this->_getBackupPlan($currentPlanId);

        $resourceId = $this->_getResourceId();
        $BackupPlan->removeResource($resourceId);
    }

    /**
     * Await api update - should update to not protected
     * when plan is revoked.
     */
    private function _awaitApiUpdateRevokePlan()
    {
        $attempts = 0;

        while ($this->_isProtected() && $attempts < 20) {
            $attempts++;
            sleep(1);
        }
    }

    /**
     * List name and id of compatible backup plans. Does not include selected plans.
     */
    public function listCompatiblePlansAction()
    {
        $plans = $this->_getApplicablePlans();
        $this->_displayPayloadAsJson($plans);
    }

    /**
     * Activate interworx default plan. Will revoke current plan.
     *
     * @param Form_SubmitOnly $Form
     */
    public function activateInterWorxDefaultPlanCommitAction(Form_SubmitOnly $Form)
    {
        if ($this->_isProtected()) {
            $this->_removeResourceFromPlan();
        }

        try {
            $this->_addResourceToExistingIworxPlan();
        } catch (Exception $e) {
            $this->_createNewPlan();
        }

        $this->_displayPayloadActivatePlan();
    }

    /**
     * Add resource to existing plan.
     */
    private function _addResourceToExistingIworxPlan()
    {
        $planId = $this->_getDefaultPlanIdFromExisting();
        $this->_addResourceToPlan($planId);
    }

    /**
     * @return string
     */
    private function _getDefaultPlanIdFromExisting()
    {
        $plans = $this->_getApplicablePlans();

        foreach ($plans as $plan) {
            $id = $plan['ID'];
            $BackupPlan = new Model_Backups_BackupPlan($id);
            if ($BackupPlan->getName() === Model_Backups_Api::DEFAULT_NAME && $BackupPlan->prePostScriptsOK()) {
                return $id;
            }
        }

        throw new Exception('no default plan was found in availabla plans');
    }

    /**
     * @return array
     */
    private function _getApplicablePlans()
    {
        $Server = $this->_getServer();
        $plans = $Server->getApplicablePlans();
        return $plans;
    }

    /**
     * Activate backup plan.
     *
     * @param Form_AbActivatePlan $Form
     */
    public function activatePlanCommitAction(Form_AbActivatePlan $Form)
    {
        $this->_executeFunctionAndCatchException('_activatePlan', $Form);
    }

    /**
     * @param Form_AbActivatePlan $Form
     */
    private function _activatePlan(Form_AbActivatePlan $Form)
    {
        if ($this->_isProtected()) {
            $this->_replaceCurrentPlan($Form->getSelectedPlanId());
        } else {
            $this->_applyOrCreateBackupPlan($Form);
            $this->_awaitApiUpdateNextBackup();
        }

        $this->_displayPayloadActivatePlan();
    }

    /**
     * check if resource is protected by backup plan.
     *
     * @return boolean
     */
    private function _isProtected()
    {
        $Server = $this->_getServer();
        return $Server->isProtected();
    }

    /**
     * switch backup plan.
     *
     * @param string $newPlanId
     */
    private function _replaceCurrentPlan($newPlanId)
    {
        $this->_removeResourceFromPlan();
        $this->_addResourceToPlan($newPlanId);
    }



    /**
     * @param string $planId
     */
    private function _addResourceToPlan($planId)
    {
        $BackupPlan = $this->_getBackupPlan($planId);
        $resourceId = $this->_getResourceId();
        $BackupPlan->addResource($resourceId);
    }

    /**
     * Get current plan id.
     */
    private function _getCurrentPlanId()
    {
        return $this->_getServer()->getPlanId();
    }

    /**
     * display successful payload after activatePlan.
     */
    private function _displayPayloadActivatePlan()
    {
        $BackupPlan = $this->_getCurrentBackupPlan();
        $payloadData = $BackupPlan->getPlanDetails();
        $payloadData['message'] = 'Backup plan successfully activated';
        $this->_displayPayloadAsJson($payloadData);
    }

    /**
     * Attempt to activate existing plan.
     * Create new plan if existing plans are invalid,
     * or if none exists.
     *
     * @param Form_AbActivatePlan $Form
     */
    private function _applyOrCreateBackupPlan(Form_AbActivatePlan $Form)
    {
        try {
            $planId = $Form->getPlanId();
            $this->_addResourceToPlan($planId);
        } catch (Exception $e) {
            // No compatible plans exists. Create new one.
            $this->_createNewPlan();
        }
    }

    /**
     * @param string $id
     * @return Model_Backups_BackupPlan
     */
    private function _getBackupPlan($id)
    {
        return new Model_Backups_BackupPlan($id);
    }

    /**
     * Create new default plan.
     */
    private function _createNewPlan()
    {
        $this->_getAcronisPlugin()->getNewPlan();
    }

    /**
     * @return string
     */
    private function _getResourceId()
    {
        return $this->_getAcronisPlugin()->getServer()->getResourceId();
    }

    /**
     * Await backup status change to 'running'.
     */
    private function _awaitApiUpdateNextBackup()
    {
        $Server = $this->_getServer();
        $attempts = 0;

        while ($Server->backupNotScheduled() && $attempts < 10) {
            sleep(2);
            $attempts++;
        }
    }

    /**
     * Get alerts.
     *
     * @param Form_SubmitOnly $Form
     */
    public function getAlertsCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_getAlerts', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _getAlerts(Form_SubmitOnly $Form)
    {
        $AlertsManager = $this->_Plugin->getAlertManager();
        $alerts = $AlertsManager->getAlerts();
        $this->_displayPayloadAsJson($alerts);
    }

    /**
     * Run backup.
     *
     * @param Form_SubmitOnly $Form
     */
    public function runBackupCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_runBackup', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _runBackup(Form_SubmitOnly $Form)
    {
        $BackupPlan = $this->_getCurrentBackupPlan();
        $resourceId = $this->_getResourceId();
        $BackupPlan->runBackup($resourceId);
        $this->_awaitResourceBackupStatus();

        $backupInfo = $this->_getBackupStatusFromResource();
        $backupInfo['message'] = 'backup successfully started'; // TODO add to lang file
        $this->_displayPayloadAsJson($backupInfo);
    }

    /**
     * Await backup status change to 'running'.
     */
    private function _awaitResourceBackupStatus()
    {
        $Server = $this->_getServer();
        $attempts = 0;

        while (!$Server->backupInProgress() && $attempts < 6) {
            sleep(10);
            $attempts++;
        }
    }

    /**
     * Get backups.
     *
     * @param Form_SubmitOnly $Form
     */
    public function listBackupsCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_listBackups', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _listBackups(Form_SubmitOnly $Form)
    {
        $backups = array();

        if ($this->_isProtected()) {
            $BackupsList = $this->_getBackupsList();
            $backups = $BackupsList->getBackups();
        }

        $this->_displayPayloadAsJson($backups);
    }

    /**
     * @return Model_Backups_List_Nodeworx
     */
    private function _getBackupsList()
    {
        $instanceId = $this->_getServer()->getInstanceId();
        return new Model_Backups_List_Nodeworx($instanceId);
    }


    /**
     * @return Plugin_Acronis
     */
    private function _getAcronisPlugin()
    {
        if (!isset($this->_Plugin)) {
            $this->_Plugin = $this->_getPlugin();
        }

        return $this->_Plugin;
    }

    /**
     * Get root content.
     *
     * @param Form_AcronisBackups $Form
     */
    public function getRootContentCommitAction(Form_AcronisBackups $Form)
    {
        $this->_executeFunctionAndCatchException('_getRootContent', $Form);
    }

    /**
     * @param Form_AcronisBackups $Form
     */
    private function _getRootContent(Form_AcronisBackups $Form)
    {
        $backupId = $Form->getBackupId();
        $FilesRestorer = $this->_Plugin->getNwFilesRecoveryPoint($backupId);
        $dirContent = $FilesRestorer->getRootContent();

        $this->_displayPayloadAsJson($dirContent);
    }

    /**
     * Get directory content by id.
     *
     * @param Form_AcronisBackups_ListDirectoryContent $Form
     */
    public function getDirectoryContentCommitAction(Form_AcronisBackups_ListDirectoryContent $Form)
    {
        $this->_executeFunctionAndCatchException('_getDirectoryContent', $Form);
    }

    /**
     * @param Form_AcronisBackups_ListDirectoryContent $Form
     */
    private function _getDirectoryContent(Form_AcronisBackups_ListDirectoryContent $Form)
    {
        $backupId = $Form->getBackupId();
        $dirPath = $Form->getDirectoryPath();

        $FilesRestorer = $this->_getAcronisPlugin()->getNwFilesRecoveryPoint($backupId);
        $dirContent = $FilesRestorer->getDirectoryContent($dirPath);

        $this->_displayPayloadAsJson($dirContent);
    }


    /**
     * Recover directories.
     *
     * @param Form_AcronisBackups_RecoverDirectories $Form
     */
    public function recoverDirectoriesCommitAction(Form_AcronisBackups_RecoverDirectories $Form)
    {
        $this->_executeFunctionAndCatchException('_recoverDirectories', $Form);
    }

    /**
     * @param Form_AcronisBackups_RecoverDirectories $Form
     */
    private function _recoverDirectories(Form_AcronisBackups_RecoverDirectories $Form)
    {
        $this->_validateNoRestoresInitiated();
        $this->_validateNoActiveDatabaseRestores();
        $this->_validateNoActiveFileRestores();

        $backupId = $Form->getBackupId();
        $dirPaths = $Form->getDirectoryPaths();
        $shouldDeleteNewFiles = $Form->deleteRequested();

        $this->_recoverDirectoriesAndHandleInitiated($backupId, $dirPaths, $shouldDeleteNewFiles);
    }

    /**
     * set restore initiated & recover directories. Clear initiated after
     * success or if error occurs.
     *
     * @param string  $backupId
     * @param array   $dirPaths
     * @param boolean $shouldDeleteNewFiles
     */
    private function _recoverDirectoriesAndHandleInitiated($backupId, array $dirPaths, $shouldDeleteNewFiles)
    {
        try {
            Model_Recovery_Files::setRestoreInitiated();
            $result = $this->_recoverDirectoriesAwaitApi($backupId, $dirPaths, $shouldDeleteNewFiles);
            Model_Recovery_Files::clearRestoreInitiated();

            $this->_displayPayloadAsJson($result);
        } catch (Exception $e) {
            Model_Recovery_Files::clearRestoreInitiated();
            throw $e;
        }
    }

    /**
     * Recover directories and wait until acronis api includes recovery.
     *
     * @param string  $backupId
     * @param array   $dirPaths
     * @param boolean $shouldDeleteNewFiles
     */
    private function _recoverDirectoriesAwaitApi($backupId, array $dirPaths, $shouldDeleteNewFiles)
    {
        $RecoveryPoint = new Model_Recovery_Files_Nodeworx($backupId);
        $RecoveryPoint->recoverDirectories($dirPaths, $shouldDeleteNewFiles);

        $planId = $RecoveryPoint->getPlanId();
        $this->_awaitRecoveryApiUpdate($planId);

        return array(
            'message' => AcronisReply::LANG_IN_PROGRESS,
            'planId' => $planId,
            'directories' => $dirPaths
        );
    }

    /**
     * Throw exception if restore is initiated.
     */
    private function _validateNoRestoresInitiated()
    {
        if (Model_Recovery_Files::restoreInitiated()) {
            throw new Exception_Acronis_RestoreInProgress();
        }
    }

    /**
     * Throws exception if restores in status "started" exists.
     */
    private function _validateNoActiveFileRestores()
    {
        $RecoveryPlans = new Model_RecoveryPlan_Plans();
        $plans = $RecoveryPlans->getAll();

        foreach ($plans as $plan) {
            if ($this->_fileRestoreInProgress($plan)) {
                throw new Exception_Acronis_RestoreInProgress();
            }
        }
    }

    /**
     *
     */
    public function recoveryInProgressAction()
    {
        try {
            $this->_validateNoActiveFileRestores();
            $this->_validateNoActiveDatabaseRestores();
            $this->_validateNoRestoresInitiated();

            $inProgress = false;
        } catch (Exception_Acronis $e) {
            $inProgress = true;
        } finally {
            $this->_displayPayloadAsJson(array(
                'recoveryInProgress' => $inProgress
            ));
        }
    }

    /**
     * Check if plan is in state started or enqueued.
     *
     * @param array $plan
     * @return boolean
     */
    private function _fileRestoreInProgress(array $plan)
    {
        $state = $plan['state'];
        return $state === 'started' || $state === 'enqueued';
    }

    /**
     * Make sure no active db restores in cache.
     */
    private function _validateNoActiveDatabaseRestores()
    {
        $Cache = new Model_Cache_Restore_Databases();

        if ($Cache->restoresInProgress()) {
            throw new Exception_Acronis_RestoreInProgress();
        }
    }

    /**
     * Await acronis api to update - including plan id.
     *
     * @param string $planId
     */
    private function _awaitRecoveryApiUpdate($planId)
    {
        $attempts = 0;
        while (!$this->_planExistsInApi($planId) && $attempts < 20) {
            $attempts++;
            sleep(3);
        }
    }

    /**
     * Check if recovery plan exists in acronis api.
     *
     * @param string $planId
     */
    private function _planExistsInApi($planId)
    {
        try {
            // will throw exception if planid not found in api
            $RecoveryPlan = new Model_RecoveryPlan_Plan($planId);
            return true;
        } catch (Exception $e) {
            return false;
        }
    }

    /**
     * Check recovery state & progress by recovery plan id.
     *
     * @param Form_AcronisBackups_RecoveryProgress $Form
     */
    public function getRecoveryProgressAction(Form_AcronisBackups_RecoveryProgress $Form)
    {
        $this->_executeFunctionAndCatchException('_getRecoveryProgress', $Form);
    }

    /**
     * Check recovery state by recovery plan id.
     *
     * @param Form_AcronisBackups_RecoveryProgress $Form
     */
    private function _getRecoveryProgress(Form_AcronisBackups_RecoveryProgress $Form)
    {
        $planId = $Form->getPlanId();
        $RecoveryPlan = new Model_RecoveryPlan_Plan($planId);

        $result = array(
            'progress' => $RecoveryPlan->getProgress(),
            'state' => $RecoveryPlan->getState(),
            'codeType' => $RecoveryPlan->getCodeType(),
            'codeError' => $RecoveryPlan->getCodeError(),
            'recoveryId' => $RecoveryPlan->getUuid(),
            'planId' => $RecoveryPlan->getPlanId(),
            'recoveryError' => $RecoveryPlan->getErrorMessage()
        );

        $this->_displayPayloadAsJson($result);
    }

    /**
     * Get recovery error by plan id.
     *
     * @param Form_AcronisBackups_RecoveryProgress $Form
     */
    public function getRecoveryErrorAction(Form_AcronisBackups_RecoveryProgress $Form)
    {
        // errors can be very bloated - message is usually enough which is included
        // in getRecoveryProgress. This action is for further troubleshooting.
        $this->_executeFunctionAndCatchException('_getRecoveryError', $Form);
    }

    /**
     * @param Form_AcronisBackups_RecoveryProgress $Form
     */
    private function _getRecoveryError(Form_AcronisBackups_RecoveryProgress $Form)
    {
        $planId = $Form->getPlanId();
        $RecoveryPlan = new Model_RecoveryPlan_Plan($planId);

        $result = array(
            'progress' => $RecoveryPlan->getProgress(),
            'state' => $RecoveryPlan->getState(),
            'codeType' => $RecoveryPlan->getCodeType(),
            'codeError' => $RecoveryPlan->getCodeError(),
            'recoveryId' => $RecoveryPlan->getUuid(),
            'planId' => $RecoveryPlan->getPlanId(),
            'recoveryError' => $RecoveryPlan->getError()
        );

        $this->_displayPayloadAsJson($result);
    }


    /**
     * List recoveries.
     *
     * @param Form_SubmitOnly $Form
     */
    public function listRecoveriesCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_listRecoveries', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _listRecoveries(Form_SubmitOnly $Form)
    {
        $RecoveryPlans = new Model_RecoveryPlan_Plans();
        $plans = $RecoveryPlans->getAll();
        $this->_displayPayloadAsJson($plans);
    }

    /**
     * Resume recovery if state is "paused".
     *
     * @param Form_AcronisBackups_ResumeRecovery $Form
     */
    public function resumeRecoveryAction(Form_AcronisBackups_ResumeRecovery $Form)
    {
        $this->_executeFunctionAndCatchException('_resumeRecovery', $Form);
    }

    /**
     * Resume recovery if state is "paused".
     *
     * @param Form_AcronisBackups_ResumeRecovery $Form
     */
    private function _resumeRecovery(Form_AcronisBackups_ResumeRecovery $Form)
    {
        $recoveryId = $Form->getRecoveryId();
        $RecoveryPlans = new Model_RecoveryPlan_Plans();
        $RecoveryPlans->resumeRecovery($recoveryId);

        $this->_displayPayloadAsJson(array(
            'message' => 'recovery resumed',
            'recoveryId' => $recoveryId
        ));
    }

    /**
     * Get restore location.
     */
    public function getDefaultRestoreLocationAction()
    {
        $altLocation = Model_Recovery_Files_Nodeworx::getDefaultRestorePath();
        $PayloadData = array(
            'location' => $altLocation
        );

        $this->_displayPayloadAsJson($PayloadData);
    }

    /**
     * Recover directories to alt location.
     *
     * @param Form_AcronisBackups_RestoreAltLocation $Form
     */
    public function restoreToAltLocationCommitAction(Form_AcronisBackups_RestoreAltLocation $Form)
    {
        $this->_executeFunctionAndCatchException('_restoreToAltLocation', $Form);
    }

    /**
     * @param Form_AcronisBackups_RestoreAltLocation $Form
     */
    private function _restoreToAltLocation(Form_AcronisBackups_RestoreAltLocation $Form)
    {
        $this->_validateNoRestoresInitiated();
        $this->_validateNoActiveDatabaseRestores();
        $this->_validateNoActiveFileRestores();

        $backupId = $Form->getBackupId();
        $dirPaths = $Form->getDirectoryPaths();

        $this->_restoreToAltLocationAndHandleInitiated($backupId, $dirPaths);
    }

    /**
     * Handle initiated flag - set initiated before attempt
     * and clear initiated when done / if error.
     *
     * @param string $backupId
     * @param array  $dirPaths
     */
    private function _restoreToAltLocationAndHandleInitiated($backupId, array $dirPaths)
    {
        try {
            Model_Recovery_Files::setRestoreInitiated();
            $result = $this->_restoreAltLocationAwaitApiUpdate($backupId, $dirPaths);
            Model_Recovery_Files::clearRestoreInitiated();

            $this->_displayPayloadAsJson($result);
        } catch (Exception $e) {
            Model_Recovery_Files::clearRestoreInitiated();
            throw $e;
        }
    }

    /**
     * Restore to alt location & return result when acronis api
     * has been updated to include restore.
     *
     * @param string $backupId
     * @param array  $dirPaths
     * @return array
     */
    private function _restoreAltLocationAwaitApiUpdate($backupId, array $dirPaths)
    {
        $FilesRestorer = $this->_getFileRestorer($backupId);
        $FilesRestorer->restoreToAltLocation($dirPaths);
        $planId = $FilesRestorer->getPlanId();
        $fullAltLocation = $FilesRestorer->getRestoreLocation();

        $this->_getActivitiesLogger()->logDownloadFiles($dirPaths);
        $this->_awaitRecoveryApiUpdate($planId);

        return array(
            'message' => AcronisReply::LANG_IN_PROGRESS_ALT_LOCATION . " $fullAltLocation",
            'planId' => $planId,
            'directories' => $dirPaths,
            'location' => $fullAltLocation
        );
    }

    /**
     * @param string $backupId
     * @return Model_Recovery_Files_Nodeworx
     */
    private function _getFileRestorer($backupId)
    {
        return $this->_Plugin->getNwFilesRecoveryPoint($backupId);
    }

    /**
     * List databases.
     *
     * @param Form_AcronisBackups $Form
     */
    public function listDatabasesCommitAction(Form_AcronisBackups $Form)
    {
        $this->_executeFunctionAndCatchException('_listDatabases', $Form);
    }

    /**
     * @param Form_AcronisBackups $Form
     */
    private function _listDatabases(Form_AcronisBackups $Form)
    {
        $backupId = $Form->getBackupId();
        $DbRestorer = $this->_Plugin->getNwDatabaseRecoveryPoint($backupId);
        $dbs = $DbRestorer->getDbs();
        $this->_displayPayloadAsJson($dbs);
    }

    /**
     * Restore datbases.
     *
     * @param Form_AcronisBackups_RecoverDatabases $Form
     */
    public function restoreDatabasesCommitAction(Form_AcronisBackups_RecoverDatabases $Form)
    {
        $this->_executeFunctionAndCatchException('_restoreDatabases', $Form);
    }

    /**
     * @param Form_AcronisBackups_RecoverDatabases $Form
     */
    private function _restoreDatabases(Form_AcronisBackups_RecoverDatabases $Form)
    {
        $this->_validateNoActiveDatabaseRestores();

        $backupId = $Form->getBackupId();
        $databases = $Form->getDatabases();
        $restoreId = IWorxSecUtil::getUuid();
        $unmapped = $Form->isUnmapped();


        $payloadData = array(
            AcronisReply::REPLY_MESSAGE => 'Database recovery started',
            AcronisReply::REPLY_DATABASES => $databases,
            AcronisReply::REPLY_RESTORE_ID => $restoreId,
            AcronisReply::REPLY_UNMAPPED => $unmapped
        );

        if ($unmapped) {
            $this->_recoverDbsInBackground($backupId, $databases, $restoreId);
        } else {
            $unixUser = $Form->getUnixUser();
            $this->_recoverDbsInBackground($backupId, $databases, $restoreId, $unixUser);

            $payloadData['unixuser'] = $unixUser;
        }

        $this->_getActivitiesLogger()->logRestoreDatabases($databases);
        $this->_displayPayloadAsJson($payloadData);
    }

    /**
     * @param string $backupId
     * @param array  $databases
     * @param string $id
     * @param string $unixUser
     */
    private function _recoverDbsInBackground($backupId, array $databases, $id, $unixUser = null)
    {
        $unmapped = is_null($unixUser);
        $functionToCall = $unmapped ? '--restoreUnmappedDatabases' : '--restoreSiteWorxDatabases';

        $phpbin = INI::get(INI::IWORX_BIN, 'php');
        $dbRestorerBin = INI::get(INI::IWORX_PLUGINS, 'dir') . '/acronis/bin/acronisdbrestorer.php';

        $dbs = implode(",", $databases);
        $dbsArg = "--databases='$dbs'";
        $backupIdArg = "--backup_id='$backupId'";
        $userArg = $unmapped ? '' : "--unixuser='$unixUser'";
        $idArg = "--restore_id='$id'";
        $cmd = "$phpbin $dbRestorerBin $functionToCall $dbsArg $userArg $idArg $backupIdArg > /dev/null 2>&1 &";

        exec($cmd);
        IWorxLog::log("Acronis: Running db restore script in background: $dbRestorerBin");
    }

    /**
     * Get database recovery state.
     *
     * @param Form_AcronisBackups_DatabaseRecoveryState $Form
     */
    public function getDatabaseRecoveryStateAction(Form_AcronisBackups_DatabaseRecoveryState $Form)
    {
        $this->_executeFunctionAndCatchException('_getDatabaseRecoveryState', $Form);
    }

    /**
     * @param Form_AcronisBackups_DatabaseRecoveryState $Form
     */
    private function _getDatabaseRecoveryState(Form_AcronisBackups_DatabaseRecoveryState $Form)
    {
        $id = $Form->getId();

        $Cache = new Model_Cache_Restore_Databases();
        $restore = $Cache->get($id);

        if (is_null($restore)) {
            throw new Exception_Acronis_InvalidInput($id);
        }

        $this->_displayPayloadAsJson($restore);
    }


    /**
     * List database recoveries
     */
    public function listDatabaseRecoveriesAction()
    {
        $Cache = new Model_Cache_Restore_Databases();
        $restores = $Cache->getAll();
        $this->_displayPayloadAsJson($restores);
    }

    /**
     * Restore Databases to alt location.
     *
     * @param Form_AcronisBackups_RecoverDatabases $Form
     */
    public function restoreDatabasesAltLocationCommitAction(Form_AcronisBackups_RecoverDatabases $Form)
    {
        $this->_executeFunctionAndCatchException('_restoreDatabasesAltLocation', $Form);
    }

    /**
     * @param Form_AcronisBackups_RecoverDatabases $Form
     */
    private function _restoreDatabasesAltLocation(Form_AcronisBackups_RecoverDatabases $Form)
    {
        $backupId = $Form->getBackupId();
        $unixUser = $Form->getUnixUser();
        $databases = $Form->getDatabases();

        $this->_validateDbsExist($backupId, $databases, $unixUser);
        $this->_validateNoActiveDatabaseRestores();
        $this->_validateNoActiveFileRestores();
        $this->_validateNoRestoresInitiated();

        $DbRestorer = $this->_getDbRestorer($backupId);
        $Backup = $this->_getBackup($backupId);
        $restoreDirName = is_null($unixUser) ? $Backup->getUnmappedDbFolder() : $unixUser;

        $DbRestorer->restoreDatabasesAltLocation($restoreDirName, $databases);
        $fullAltLocation = $DbRestorer->getRestoreLocation();
        $planId = $DbRestorer->getPlanId();
        $this->_getActivitiesLogger()->logDownloadDatabases($databases);

        $payloadData = array(
            'message' => 'Database files successfully recovered',
            'unixuser' => $unixUser,
            'databases' => $databases,
            'location' => $fullAltLocation,
            'planId' => $planId
        );

        $this->_displayPayloadAsJson($payloadData);
    }

    /**
     * @param string $backupId
     * @param array  $databases
     * @param string $unixuser
     */
    private function _validateDbsExist($backupId, array $databases, $unixuser = null)
    {
        $Backup = $this->_getBackup($backupId);
        $dbsBackup = is_null($unixuser) ? $Backup->getUnmappedDatabases() : $Backup->getUserDatabases($unixuser);

        if (empty($dbsBackup)) {
            throw new Exception_Acronis_DbNotFound($databases);
        }

        foreach ($databases as $dbToRestore) {
            if (!$this->_dbExists($dbsBackup, $dbToRestore)) {
                throw new Exception_Acronis_DbNotFound($dbToRestore);
            }
        }
    }

    /**
     * @param array  $dbsBackup
     * @param string $dbToRestore
     */
    private function _dbExists(array $dbsBackup, $dbToRestore)
    {
        foreach ($dbsBackup as $dbBackup) {
            if ($dbBackup['name'] === $dbToRestore) {
                return true;
            }
        }

        return false;
    }

    /**
     * @param string $backupId
     * @return Model_Recovery_Files_Databases_Nodeworx|Mock
     */
    private function _getDbRestorer($backupId)
    {
        return $this->_Plugin->getNwDatabaseRecoveryPoint($backupId);
    }

    /**
     * @param string $backupId
     * @return Model_Backups_Backup
     */
    private function _getBackup($backupId)
    {
        $cache = new Model_Cache_Backup();
        $backup = $cache->getApiData($backupId);
        return new Model_Backups_Backup($backup);
    }

    /**
     * Remove any unmapped prescripts - use with caution.
     *
     * @param Form_SubmitOnly $Form
     */
    public function cleanupUnmappedPrescriptDataCommitAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_cleanupUnmappedPrescriptData', $Form);
    }

    /**
     * Remove any unmapped prescripts - use with caution.
     *
     * @param Form_SubmitOnly $Form
     */
    public function _cleanupUnmappedPrescriptData(Form_SubmitOnly $Form)
    {
        $Plugin = $this->_getAcronisPlugin();
        $BackupsList = $Plugin->getNwBackupsList();
        $BackupsList->deleteUnmappedPrescripts();

        $payloadData = array(
            'message' => 'Unmapped prescript data successfully deleted'
        );

        $this->_displayPayloadAsJson($payloadData);
    }

    /**
     * List unmapped prescript data
     *
     * @param Form_SubmitOnly $Form
     */
    public function listUnmappedPrescriptDataAction(Form_SubmitOnly $Form)
    {
        $this->_executeFunctionAndCatchException('_listUnmappedPrescriptData', $Form);
    }

    /**
     * @param Form_SubmitOnly $Form
     */
    private function _listUnmappedPrescriptData(Form_SubmitOnly $Form)
    {
        $Plugin = $this->_getAcronisPlugin();
        $BackupsList = $Plugin->getNwBackupsList();

        $prescriptData = $BackupsList->getUnmappedPrescriptData();
        $this->_displayPayloadAsJson($prescriptData);
    }

    /**
     * Delete backups from cache after x days
     *
     * @param Form_AcronisBackups_ExpireBackup $Form
     */
    public function setBackupExpirationCommitAction(Form_AcronisBackups_ExpireBackup $Form)
    {
        $this->_executeFunctionAndCatchException('_setBackupExpiration', $Form);
    }

    /**
     * @param Form_AcronisBackups_ExpireBackup $Form
     */
    private function _setBackupExpiration(Form_AcronisBackups_ExpireBackup $Form)
    {
        $days = $Form->getDays();
        $Cache = new Model_Cache_Backup();
        $Cache->setExpireAfterDays($days);

        $payloadData = AcronisReply_Backup::getExpireBackupResponse($days);
        $this->_displayPayloadAsJson($payloadData);
    }



    /**
     * @param string $functionToCall
     * @param Form   $Form
     */
    private function _executeFunctionAndCatchException($functionToCall, Form $Form)
    {
        try {
            $this->$functionToCall($Form);
        } catch (Exception_Acronis $e) {
            IW::getLog()->error("Action failed $functionToCall");
            $e->logError();

            $errorCode = $e->getErrorCode();
            $langKey = AcronisReply::getErrorLangKey($errorCode);
            $this->_displayFailure('index', $langKey, $errorCode);
        } catch (Exception $e) {
            IW::getLog()->error("Action failed $functionToCall");
            IWorxLog::log("Acronis: " . $e->getMessage(), IWorxLog::WARN);
            $this->_displayFailure('index', $e->getMessage(), Exception_Acronis::getDefaultErrorCode());
        }
    }

    /**
     * Display json if cli
     *
     * @param mixed $payloadContent
     */
    private function _displayPayloadAsJson($payloadContent)
    {
        // render as json
        if (IW::Env()->matches(IW::CLI)) {
            IW::Env()->setOutputStyle('pretty');
        }
        $Payload = new Payload($payloadContent);
        $this->_displayPayload($Payload);
    }

    /**
     * get nw activities logger.
     *
     * @return Model_Logger_Activities_Nodeworx
     */
    protected function _getActivitiesLogger()
    {
        if (!isset($this->_ActivitiesLogger)) {
            $this->_ActivitiesLogger = new Model_Logger_Activities_Nodeworx();
        }

        return $this->_ActivitiesLogger;
    }
}
