<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Malin Lorch <mlorch@liquidweb.com>
 * @date    Nov 27, 2020
 */

/**
 * Class that determines which device path resolves to.
 *
 * @package Interworx
 * @subpackage Plugin_Acronis
 */
class Model_Directory_FileLocator
{
    private $_path;
    private $_deviceName;

    /**
     * Expects local path.
     *
     * @param string $path
     */
    public function __construct($path)
    {
        $this->_path = '/' . trim($path, '/');

        if ($this->pathStartsWith('/chroot')) {
            $this->_deviceName = $this->_getDeviceName('/chroot');
            $this->_replaceWithRealPath('/chroot');
        }

        if ($this->pathStartsWith('/home')) {
            $this->_deviceName = $this->_getDeviceName('/home');
            $this->_replaceWithRealPath('/home');
        }

        if (!isset($this->_deviceName)) {
            $this->_deviceName = $this->getDefaultDeviceName();
        }
    }

    /**
     * @param string $dirToReplace
     */
    private function _replaceWithRealPath($dirToReplace)
    {
        $realpath = $this->_getRealPath($dirToReplace);

        if ($realpath !== $dirToReplace) {
            $this->_path = substr_replace($this->_path, $realpath, 0, strlen($dirToReplace));
        }
    }

    /**
     * Get realpath. Need to execute as root, built-in php function wont work.
     *
     * @param string $path
     * @return string
     */
    private function _getRealPath($path)
    {
        $Command = new Command("realpath $path");
        $Command->runAsRoot();
        $Command->exec();
        $realpath = $Command->getOutput()[0];

        if (!$Command->wasSuccessful()) {
            $Command->logResult(true);
            throw new Exception_Acronis_InvalidPath($path);
        }

        return $realpath;
    }

    /**
     * Get path (local path)
     *
     * @return string
     */
    public function getPath()
    {
        return $this->_path;
    }

    /**
     * check if path (haystack) starts with given needle.
     *
     * @param string $startsWith
     * @return boolean
     */
    public function pathStartsWith($startsWith)
    {
        $pos = strpos($this->_path, $startsWith);
        return $pos !== false && ($pos === 0 || $pos === 1);
    }

    /**
     * Get device name
     *
     * @param string $path
     * @return string
     */
    private function _getDeviceName($path)
    {
        $path = '/' . ltrim($path, '/');
        $Command = new Command("df -P $path | awk 'NR==1 {next} {print $1; exit}'");
        $Command->allowPipe();
        $Command->allowQuoteSingle();
        $Command->allowCurlyBoth();
        $Command->allowSemicolon();
        $Command->allowDollarSign();
        $Command->exec();

        if (!$Command->wasSuccessful()) {
            throw new Exception_Acronis_InvalidPath($path);
        }

        $path = $Command->getOutput()[0];
        $deviceName = end(explode('/', $path));
        return $deviceName;
    }

    /**
     * Check if mp is default (/).
     *
     * @return boolean
     */
    public function getDefaultDeviceName()
    {
        return $this->_getDeviceName('/');
    }

    /**
     * Check if default device.
     *
     * @return boolean
     */
    public function onDefaultDevice()
    {
        return $this->_deviceName === $this->getDefaultDeviceName();
    }

    /**
     * Get device name that path resolves to.
     *
     * @return string
     */
    public function getDeviceName()
    {
        return $this->_deviceName;
    }

    /**
     * Get root device path (according to acronis api).
     *
     * @return string
     */
    public function getRootDevicePath()
    {
        return "$this->_deviceName:";
    }

    /**
     * Check if path is chroot.
     *
     * @param string $pathToCompare
     * @return boolean
     */
    public function pathEquals($pathToCompare)
    {
        return $this->_path === $pathToCompare || $this->_path === "/$pathToCompare";
    }

    /**
     * Get device path excluding device name.
     *
     * @return string
     */
    public function getDevicePath()
    {
        if ($this->onDefaultDevice()) {
            return $this->_path;
        }

        return $this->_getNonDefaultDevicePath();
    }

    /**
     * Get full device path. Includes device name.
     *
     * @return string
     */
    public function getFullDevicePath()
    {
        $root = $this->getRootDevicePath();
        $devicePath = $this->getDevicePath();

        return "$root/$devicePath";
    }

    /**
     * @return string
     */
    private function _getNonDefaultDevicePath()
    {
        if ($this->onDefaultDevice()) {
            throw new Exception('on default device!');
        }

        $dirs = $this->_pathToArray($this->_path);
        $excludingMountpoints = $this->_excludeMountpoints($dirs);
        $path = implode('/', $excludingMountpoints);

        return $path;
    }

    /**
     * @param array $dirs
     * @return array
     */
    private function _excludeMountpoints(array $dirs)
    {
        // exclude possible mountpoints
        if ($dirs[0] === 'home' || $dirs[0] === 'chroot') {
            array_shift($dirs);
        }

        return $dirs;
    }

    /**
     * Get array of device dirs. Excludes root/device name.
     *
     * @return array
     */
    public function getDeviceDirs()
    {
        $devicePath = $this->getDevicePath();
        $dirs = $this->_pathToArray($devicePath);
        return $dirs;
    }

    /**
     * get path as array - each dir as separate item.
     *
     * @param string $path
     * @return array
     */
    private function _pathToArray($path)
    {
        return array_values(array_filter(explode('/', $path)));
    }
}
