<?php
/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2021 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @author     Malin Lorch <mlorch@liquidweb.com>
 * @date       Aug 30, 2021
 */
require_once(__DIR__ . '/../../Nodeworx.php');

/**
 * @package InterWorx
 * @subpackage
 */
class NwActivatePlanTest extends Acronis_NodeworxTest
{
    const PLAN_NO_PRESCRIPT = 'NOT DEFAULT'; // MAKE SURE THIS EXISTS IN CLOUD

    /**
     * test revokePlanCommitAction
     */
    public function testRevokePlan()
    {
        if ($this->_isActivated()) {
            echo __FUNCTION__;

            $Reply = $this->_sendCtrlRequest('revokePlan');
            $this->assertReplySuccessful($Reply);
            $this->_awaitRevoked();
        }
    }

    /**
     * Await revoked.
     */
    private function _awaitRevoked()
    {
        $attempts = 0;
        $maxAttempts = 10;

        while ($this->_isActivated() && $attempts < $maxAttempts) {
            sleep(10);
            $attempts++;
        }

        if ($this->_isActivated()) {
            $this->fail('AWAIT REVOKE - PLAN NOT REVOKED!');
        }
    }

    /**
     * Test listCompatiblePlansCommitAction
     */
    public function testListCompatiblePlans()
    {
        echo __FUNCTION__;


        $Reply = $this->_sendCtrlRequest('listCompatiblePlans');
        $this->assertReplySuccessful($Reply);

        $plans = $Reply->getPayloadData();
        $this->assertNotEmpty($plans);
    }

    /**
     * test activatePlan with & without id specified.
     * Expects most recent plan to be activated if revoking
     * then activating without id specified
     */
    public function testActivatePlanNoPrescript()
    {
        echo __FUNCTION__;


        // make sure this plan exists!
        $planIdNoPrescript = $this->_getPlanId(self::PLAN_NO_PRESCRIPT);
        if (is_null($planIdNoPrescript)) {
            $this->fail('plan does not exist! Create in cloud and run tests again: ' . self::PLAN_NO_PRESCRIPT);
        }

        $Reply = $this->_sendCtrlRequest('activatePlan', array(
            Form_AbActivatePlan::INPUT_PLAN => $planIdNoPrescript
        ));

        $this->assertReplySuccessful($Reply);

        $payloadData = $Reply->getPayloadData();
        $this->assertTrue($payloadData['enabled']);
        $this->assertSame($planIdNoPrescript, $payloadData['id']);
        $this->assertSame(self::PLAN_NO_PRESCRIPT, $payloadData['name']);


        // make sure revoking / activating without id specified enabled the same plan
        $ReplyRevokePlan = $this->_sendCtrlRequest('revokePlan');
        $this->assertReplySuccessful($ReplyRevokePlan);

        $ReplyActivateWithoutId = $this->_sendCtrlRequest('activatePlan');
        $this->assertReplySuccessful($ReplyActivateWithoutId);

        $activatedPlan = $this->_getActivatedPlan();
        $this->assertSame(trim($planIdNoPrescript), trim($activatedPlan['id']));
        $this->assertSame(self::PLAN_NO_PRESCRIPT, $activatedPlan['name']);
    }

    /**
     * Test activate default plan.
     */
    public function testActivateDefaultPlan()
    {
        echo __FUNCTION__;

        $Reply = $this->_sendCtrlRequest('activateInterWorxDefaultPlan');
        $this->assertReplySuccessful($Reply);

        $payloadData = $Reply->getPayloadData();
        $this->assertSame(Model_Backups_Api::DEFAULT_NAME, $payloadData['name']);

        $activatedPlan = $this->_getActivatedPlan();
        $this->assertSame(Model_Backups_Api::DEFAULT_NAME, $activatedPlan['name']);

        return $payloadData['id'];
    }


    /**
     * test getBackupPlanDetails.
     * Should get default plan after activating default plan.
     *
     * @depends testActivateDefaultPlan
     * @param string $expectedId
     */
    public function testGetBackupPlanDetails($expectedId)
    {
        echo __FUNCTION__;

        $Reply = $this->_sendCtrlRequest('getBackupPlanDetails');
        $this->assertReplySuccessful($Reply);


        $backupPlanDetails = $Reply->getPayloadData();

        $actualDays = $backupPlanDetails['days'];
        $actualName = $backupPlanDetails['name'];
        $actualId = $backupPlanDetails['id'];
        $actualEnabled = $backupPlanDetails['enabled'];

        $expectedName = Model_Backups_Api::DEFAULT_NAME;
        $expectedCountDays = 7;

        $this->assertArrayHasKey('enabled', $backupPlanDetails);
        $this->assertArrayHasKey('id', $backupPlanDetails);
        $this->assertArrayHasKey('name', $backupPlanDetails);
        $this->assertArrayHasKey('time', $backupPlanDetails);
        $this->assertArrayHasKey('days', $backupPlanDetails);

        $this->assertTrue($actualEnabled);
        $this->assertSame($expectedId, $actualId);
        $this->assertSame($expectedName, $actualName);
        $this->assertCount($expectedCountDays, $actualDays);

        $time = $backupPlanDetails['time'];
        $actualHour = $time['hour'];
        $actualMinute = $time['minute'];
        $actualSecond = $time['second'];

        $this->assertSame(2, $actualHour);
        $this->assertSame(0, $actualMinute);
        $this->assertSame(0, $actualSecond);
    }

    /**
     * @return array
     */
    private function _getActivatedPlan()
    {
        $Reply = $this->_sendCtrlRequest('getBackupPlanDetails');
        return $Reply->getPayloadData();
    }

    /**
     * @param string $planName
     */
    private function _getPlanId($planName)
    {
        $Reply = $this->_sendCtrlRequest('listCompatiblePlans');
        $this->assertReplySuccessful($Reply);

        $plans = $Reply->getPayloadData();
        foreach ($plans as $plan) {
            if ($plan['name'] === $planName) {
                return $plan['ID'];
            }
        }

        $this->fail('PLAN DOES NOT EXIST! CREATE PLAN IN CLOUD WITH NAME ' . self::PLAN_NO_PRESCRIPT . ' THEN TRY AGAIN');
    }
}
