<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2021 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @author     Malin Lorch <mlorch@liquidweb.com>
 * @date       Sep 13, 2021
 */

/**
 * @package InterWorx
 * @subpackage
 */
require_once(__DIR__ . '/../../Nodeworx.php');

/**
 * Note: This test depends on RunBackup.php runBackupTest.
 * Run test suite in plugin test folder to include dependencies.
 *
 * @package InterWorx
 * @subpackage
 */
class NwRecoverDatabasesAltLocationTest extends Acronis_NodeworxTest
{
    /**
     * Delete temp restore dir if exists.
     */
    public function setUp()
    {
        $this->_deleteTempDirIfExists();
    }

    /**
     * Delete temp restore dir if exists.
     */
    public function tearDown()
    {
        $this->_deleteTempDirIfExists();
    }


    /**
     * Test recover databases alt location.
     */
    public function testRecoverDatabasesAltLocationSw()
    {
        echo __FUNCTION__;
        $this->_awaitRestores();

        $dbsToRestore =  $this->_getSwDbNames();
        $Reply = $this->_recoverDatabasesAltLocationSw($dbsToRestore);
        $this->assertReplySuccessful($Reply);

        $payloadData = $Reply->getPayloadData();

        $planId = $payloadData['planId'];
        $location = $payloadData['location'];

        $this->_awaitRestore($planId);
        $this->_assertDbFilesRestoredToTemp($location, $dbsToRestore);

        return $planId;
    }

    /**
     * @depends testRecoverDatabasesAltLocationSw
     * @param string $planId
     */
    public function testListRecoveries($planId)
    {
        echo __FUNCTION__;

        $Reply = $this->_sendCtrlRequest('listRecoveries');
        $this->assertReplySuccessful($Reply);

        // assume latest restore is first item
        $payloadData = $Reply->getPayloadData();
        $restore = $payloadData[0];
        $this->assertSame(strtoupper($planId), strtoupper($restore['planId']));
    }


    /**
     * test recoverDatabasesAltLocation should fail when flag "initiated" is set
     */
    public function testRecoverDatabasesAltLocationShouldFailWhileInitiated()
    {
        echo __FUNCTION__;

        $this->_setInitiated();

        $Reply = $this->_recoverDatabasesAltLocationSw($this->_getSwDbNames());
        $this->assertReplyNotSuccessful($Reply, 'expected recoverDatabasesAltLocation to fail while initiated flag is set');

        $this->_clearInitiated();
    }


    /**
     * test recoverDatabasesAltLocation unmapped nonexisting db
     */
    public function testRecoverDatabasesAltLocationUnmappedShouldFailIfDbDoesNotExist()
    {
        echo __FUNCTION__;
        $this->_awaitRestores();

        $dbs =  array('dbdoesnotexist');
        $backupId = $this->_getBackup()->getBackupId();

        $args = array(
            Form_AcronisBackups::INPUT_DBS => $dbs,
            Form_AcronisBackups::INPUT_BACKUP_ID => $backupId
        );

        $Reply = $this->_sendCtrlRequest('restoreDatabasesAltLocation', $args);
        $this->assertReplyNotSuccessful($Reply, 'expected recoverDatabasesAltLocation to fail with non-existing db');

        $errors = $Reply->getErrors();
        $this->assertArrayContains($errors, '##LG_ACRONIS_NO_DB##');
    }

    /**
     * test recoverDatabasesAltLocation sw user nonexisting db
     */
    public function testRecoverDatabasesAltLocationSwShouldFailIfDbDoesNotExist()
    {
        echo __FUNCTION__;
        $this->_awaitRestores();

        $dbs =  array('dbdoesnotexist');
        $backupId = $this->_getBackup()->getBackupId();

        $args = array(
            Form_AcronisBackups::INPUT_DBS => $dbs,
            Form_AcronisBackups::INPUT_BACKUP_ID => $backupId,
            Form_AcronisBackups::INPUT_USER => self::ACRONISTESTUSER
        );

        $Reply = $this->_sendCtrlRequest('restoreDatabasesAltLocation', $args);
        $this->assertReplyNotSuccessful($Reply, 'expected recoverDatabasesAltLocation to fail for sw user with non-existing db');

        $errors = $Reply->getErrors();
        $this->assertArrayContains($errors, '##LG_ACRONIS_NO_DB##');
    }

    /**
     * @param array $dbs
     * @return Reply
     */
    private function _recoverDatabasesAltLocationSw(array $dbs)
    {
        $backupId = $this->_getBackup()->getBackupId();

        $args = array(
            Form_AcronisBackups::INPUT_DBS => $dbs,
            Form_AcronisBackups::INPUT_BACKUP_ID => $backupId,
            Form_AcronisBackups::INPUT_USER => self::ACRONISTESTUSER
        );

        return $this->_sendCtrlRequest('restoreDatabasesAltLocation', $args);
    }

    /**
     * @param array $dbs
     */
    private function _recoverDatabasesAltLocationUnmapped(array $dbs)
    {
        $args = array(
            Form_AcronisBackups::INPUT_DBS => $dbs,
            Form_AcronisBackups::INPUT_BACKUP_ID => $this->_getBackup()->getBackupId()
        );

        return $this->_sendCtrlRequest('restoreDatabasesAltLocation', $args);
    }

    /**
     * @return array
     */
    private function _getSwDbNames()
    {
        return array('acronistest_testdb1', 'acronistest_testdb2');
    }

    /**
     * @return string
     */
    public function testRecoverDatabasesAltLocationUnmappedDb()
    {
        echo __FUNCTION__;

        $dbs = $this->_getUnmappedDbNames();
        $Reply = $this->_recoverDatabasesAltLocationUnmapped($dbs);
        $this->assertReplySuccessful($Reply);

        $payloadData = $Reply->getPayloadData();
        $planId = $payloadData['planId'];
        $location = $payloadData['location'];

        $this->_awaitRestore($planId);

        $unmapped = true;
        $this->_assertDbFilesRestoredToTemp($location, $dbs, $unmapped);

        return $planId;
    }

    /**
     * @depends testRecoverDatabasesAltLocationUnmappedDb
     * @param string $planId
     */
    public function testListRecoveriesShouldIncludeUnmappedDbs($planId)
    {
        echo __FUNCTION__;

        $Reply = $this->_sendCtrlRequest('listRecoveries');
        $this->assertReplySuccessful($Reply);

        // assume latest restore is first item
        $payloadData = $Reply->getPayloadData();
        $restore = $payloadData[0];
        $this->assertSame(strtoupper($planId), strtoupper($restore['planId']));
    }
}
