<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2021 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package    InterWorx
 * @author     Malin Lorch <mlorch@liquidweb.com>
 * @date       Sep 03, 2021
 */

require_once(__DIR__ . '/../../Nodeworx.php');

/**
 * @package InterWorx
 * @subpackage
 */
class NwRecoverDirectoriesTest extends Acronis_NodeworxTest
{
    /**
     * Remove recovered dirs after tests.
     */
    public function tearDown()
    {
        $this->_deleteFlagInitiatedIfExists();

        if (Model_AcronisFileSys::isDir(self::PARENT_DIR_PATH_TEST)) {
            Model_AcronisFileSys::rm(self::PARENT_DIR_PATH_TEST);
        }
    }


    /**
     * Test recover directories. New dirs should not be deleted.
     */
    public function testRecoverDirectories()
    {
        echo __FUNCTION__;
        $this->_awaitRestores();

        $dir1 = self::PARENT_DIR_PATH_TEST . "/testdir_1";
        $dir2 = self::PARENT_DIR_PATH_TEST . "/testdir_2";
        $dirs =  array($dir1, $dir2);

        $this->assertFalse(Model_AcronisFileSys::isDir($dir1));
        $this->assertFalse(Model_AcronisFileSys::isDir($dir2));

        // create new dir in dir1 - should not be deleted unless "delete files" is set to true
        $newDir = "$dir1/new";
        Model_AcronisFileSys::createDir($newDir);

        $Reply = $this->_recoverDirectories($dirs);
        $this->assertReplySuccessful($Reply);

        $payloadData = $Reply->getPayloadData();
        $planId = $payloadData['planId'];

        $this->_awaitRestore($planId);

        $this->assertTrue(Model_AcronisFileSys::isDir($dir1));
        $this->assertTrue(Model_AcronisFileSys::isDir($dir2));

        // new dir should not have been deleted
        $this->assertTrue(Model_AcronisFileSys::isDir($newDir));

        return $planId;
    }


    /**
     * @depends testRecoverDirectories
     * @param string $planId
     */
    public function testGetRecoveryProgress($planId)
    {
        echo __FUNCTION__;

        $args = array(Form_AcronisBackups_RecoveryProgress::INPUT_PLAN_ID => $planId);
        $Reply = $this->_sendCtrlRequest('getRecoveryProgress', $args);

        $this->assertReplySuccessful($Reply);

        $payloadData = $Reply->getPayloadData();
        $this->assertArrayHasKey('progress', $payloadData);
        $this->assertArrayHasKey('state', $payloadData);
        $this->assertArrayHasKey('codeType', $payloadData);
        $this->assertArrayHasKey('codeError', $payloadData);
        $this->assertArrayHasKey('recoveryId', $payloadData);
        $this->assertArrayHasKey('planId', $payloadData);
        $this->assertArrayHasKey('recoveryError', $payloadData);
    }


    /**
     * @depends testRecoverDirectories
     * @param string $planId
     */
    public function testListRecoveries($planId)
    {
        echo __FUNCTION__;

        $Reply = $this->_sendCtrlRequest('listRecoveries');
        $this->assertReplySuccessful($Reply);

        // assume latest restore is first item
        $payloadData = $Reply->getPayloadData();
        $restore = $payloadData[0];
        $this->assertSame(strtoupper($planId), strtoupper($restore['planId']));
    }

    /**
     * Test recover directories. New dirs should be deleted.
     */
    public function testRecoverDirectoriesDeleteNewDirs()
    {
        echo __FUNCTION__;
        $this->_awaitRestores();

        $dir1 = self::PARENT_DIR_PATH_TEST . "/testdir_1";
        $dir2 = self::PARENT_DIR_PATH_TEST . "/testdir_2";
        $dirs =  array($dir1, $dir2);

        $this->assertFalse(Model_AcronisFileSys::isDir($dir1));
        $this->assertFalse(Model_AcronisFileSys::isDir($dir2));

        // create new dir in dir1 - should be deleted
        $newDir = "$dir1/new";
        Model_AcronisFileSys::createDir($newDir);

        $deleteNewContent = true;
        $Reply = $this->_recoverDirectories($dirs, $deleteNewContent);

        $payloadData = $Reply->getPayloadData();
        $planId = $payloadData['planId'];

        $this->_awaitRestore($planId);

        $this->assertTrue(Model_AcronisFileSys::isDir($dir1));
        $this->assertTrue(Model_AcronisFileSys::isDir($dir2));

        // new dir should have been deleted
        $this->assertFalse(Model_AcronisFileSys::isDir($newDir));
    }

    /**
     * test recoverDirectories should fail when flag "initiated" is set
     */
    public function testRecoverDirectoriesShouldFailWhileInitiated()
    {
        echo __FUNCTION__;

        IW::ENV()->setEnvFlag(Model_Recovery_Files::ENV_FLAG_RESTORE_INITIATED);

        $dirs =  array(self::PARENT_DIR_PATH_TEST . "/testdir_2");
        $Reply = $this->_recoverDirectories($dirs);

        $this->assertReplyNotSuccessful($Reply, 'Should not be able to recover dirs while initiated flag is set!');

        IW::ENV()->clearEnvFlag(Model_Recovery_Files::ENV_FLAG_RESTORE_INITIATED);
    }
}
